# import the necessary packages
from keras.utils import np_utils
import numpy as np
import h5py

class HDF5DatasetGenerator:

    def __init__(self, dbPath, batchSize, preprocessors=None, aug=None, binarize=True, classes=2):
        self.batchSize = batchSize
        self.preprocessors = preprocessors
        self.aug = aug
        self.binarize = binarize
        self.classes = classes
        self.db = h5py.File(dbPath)
        self.numImages = self.db["labels"].shape[0]
    
    def generator(self, passes=np.inf):
        # epochs = 0
        # while epochs < passes:
        for i in np.arange(0, self.numImages, self.batchSize):
            images = self.db["images"][i: i + self.batchSize]
            labels = self.db["labels"][i: i + self.batchSize]
            len_labels = self.db["len_labels"][i: i + self.batchSize]
            if self.binarize:
                labels = np_utils.to_categorical(labels, self.classes)
            if self.preprocessors is not None:
                procImages = []
                for image in images:
                    for p in self.preprocessors:
                        image = p.preprocess(image)
                    procImages.append(image)
                images = np.array(procImages)
            if self.aug is not None:
                (images, labels, len_labels) = next(self.aug.flow(images, labels, len_labels, batch_size=self.batchSize))
            yield (images, labels, len_labels)
            # epochs += 1
    
    def close(self):
        self.db.close()


import h5py
import os

class HDF5DatasetWriter:
    def __init__(self, data_dims, label_dims, outputPath, dataKey="images", bufSize=1000):
        if os.path.exists(outputPath):
            raise ValueError("The supplied ‘outputPath‘ already exists and cannot be overwritten. Manually delete the file before continuing.", outputPath)
        self.db = h5py.File(outputPath, "w")
        self.data = self.db.create_dataset(dataKey, data_dims, dtype="float")
        self.labels = self.db.create_dataset("labels", label_dims, dtype='int')
        self.len_labels = self.db.create_dataset("len_labels", (label_dims[0],), dtype='int')
        self.bufSize = bufSize
        self.buffer = {"data": [], "labels": [], "len_labels": []}
        self.idx = 0
    
    def add(self, rows, labels, lens):
        self.buffer["data"].extend(rows)
        self.buffer["labels"].extend(labels)
        self.buffer["len_labels"].extend(lens)
        if len(self.buffer["data"]) >= self.bufSize:
            self.flush()

    def flush(self):
        i = self.idx + len(self.buffer["data"])
        self.data[self.idx:i] = self.buffer["data"]
        self.labels[self.idx:i] = self.buffer["labels"]
        self.len_labels[self.idx:i] = self.buffer["len_labels"]
        self.idx = i
        self.buffer = {"data": [], "labels": [], "len_labels": []}
    
    def storeClassLabels(self, classLabels):
        dt = h5py.special_dtype(vlen=str)
        labelSet = self.db.create_dataset("label_names", (len(classLabels),), dtype=dt)
        labelSet[:] = classLabels
    
    def close(self):
        if len(self.buffer["data"]) > 0:
            self.flush()
        self.db.close()


import matplotlib.pyplot as plt
from sklearn.manifold import TSNE

def point_scatter(colors, X, Y, c):
    tsne = TSNE(n_components=2, random_state=0)
    X = tsne.fit_transform(X)
    plt.figure(figsize=(6, 5))
    plt.scatter(X, Y, c=c)
    plt.legend()
    plt.show()



import numpy as np
from tensorflow.keras.models import Sequential

class REPTILE(object):
    def __init__(self, model, XTrain, YTrain, XTest, YTest, batch_size=3, epochs=100, num_samples=300, num_iterations=10, mini_batch=10):
        
        #model
        self.model = model

        #XTrain
        self.XTrain = XTrain

        #YTrain
        self.YTtrain = YTrain

        #XTest
        self.XTest = XTest

        #YTest
        self.YTest = YTest

        #initialize number of tasks i.e number of tasks we need in each batch of tasks
        self.num_tasks = batch_size

        #number of samples i.e number of shots -number of data points (k) we need to have in each task
        self.num_samples = num_samples

        #number of epochs i.e training iterations
        self.epochs = epochs

        #hyperparameter for the inner loop (inner gradient update)
        self.alpha = 0.0001

        #hyperparameter for the outer loop (outer gradient update) i.e meta optimization
        self.beta = 0.0001

        #number of times we want to perform optimization
        self.num_iterations = num_iterations


        #mini btach size
        self.mini_batch = mini_batch
        
        #train start
        self.train_start = 0

        #test start
        self.test_start = 0
    
    def shuffle_array(self, X, Y):
        indices = np.arange(len(X[0]))
        np.random.shuffle(indices)
        Xout = []
        Yout = []
        for x in X:
            Xout.append(x[indices])
        for y in Y:
            Yout.append(y[indices])
        return Xout, Yout

    def sample_points(self, X, Y, start, num_samples):
        X, Y = self.shuffle_array(X, Y)
        Xout = []
        Yout = []
        for x in X:
            Xout.append(x[start: start+num_samples])
        for y in Y:
            Yout.append(y[start: start+num_samples])
        return Xout, Yout

    def mini_batch_data(self, X, Y, index):
        #sample mini batch of examples 
        X_minibatch = []
        Y_minibatch = []

        for x in X:
            X_minibatch.append(x[index:index+self.mini_batch])
        
        for y in Y:
            Y_minibatch.append(y[index:index+self.mini_batch])
        
        if (len(X_minibatch) == 1):
            X_minibatch = X_minibatch[0]
        
        if (len(Y_minibatch) == 1):
            Y_minibatch = Y_minibatch[0]

        return X_minibatch, Y_minibatch
    
    def test_data_convert (self, X, Y):
        if (len(X) == 1):
            X = X[0]
        if (len(Y) == 1):
            Y = Y[0]
        return X, Y

    #now let's get to the interesting part i.e training 
    def train(self):
        for e in range(self.epochs):
            #for each task in batch of tasks
            for task in range(self.num_tasks):
                
                #get the initial parameters of the model
                old_weights = self.model.get_weights()
                
                #sample x and y
                x_sample, y_sample = self.sample_points(self.XTrain, self.YTtrain, self.train_start, self.num_samples)

                #for some k number of iterations perform optimization on the task
                for k in range(self.num_iterations):

                    #get the minibatch x and y
                    for i in range(0, self.num_samples, self.mini_batch):

                        #sample mini batch of examples 
                        x_minibatch, y_minibatch = self.mini_batch_data(x_sample, y_sample, i)
                        self.model.train_on_batch(x_minibatch, y_minibatch)
                
                #get the updated model parameters after several iterations of optimization
                new_weights = self.model.get_weights()

                #Now we perform meta update

                #i.e theta = theta + epsilon * (theta_star - theta)

                epsilon = 0.01
                
                updated_weights = []
                for i in range(len(old_weights)):
                    updated_weight = old_weights[i] + epsilon * (new_weights[i] - old_weights[i])
                    updated_weights.append(updated_weight)
                
                for i in range(len(updated_weights)):
                    self.model.weights[i] = updated_weight[i]

                self.train_start += 1
                if (self.train_start + self.num_samples > len(self.XTrain)):
                    self.train_start = 0

            loss = self.model.train_on_batch(x_sample, y_sample)
            x_test, y_test = self.test_data_convert(self.XTest, self.YTest)
            score = self.model.evaluate(x_test, y_test, verbose=0)
            print ("Epoch {}: Loss {}\n".format(e,loss))
            print ("Epoch {}: score {}\n".format(e, score))
            print ('Updated Model Parameter Theta\n')
            print ('Sampling Next Batch of Tasks \n')
            print ('---------------------------------\n')
        x_test, y_test = self.test_data_convert(self.XTest, self.YTest)
        return self.model.evaluate(x_test, y_test, verbose=0)


from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import MaxPooling2D
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import Dense
from tensorflow.keras import backend as K


class LeNet:
    @staticmethod
    def build(width, height, depth, classes, last_active="softmax"):
        # Initialize the model
        model = Sequential()
        input_shape = (height, width, depth)

        # If we are using 'channels-first', update the input shape
        if K.image_data_format() == 'channels_first':
            input_shape = (depth, height, width)

        # First set of CONV => RELU => POOL layers
        model.add(Conv2D(20, (5, 5), padding='same', input_shape=input_shape))
        model.add(Activation('relu'))
        model.add(MaxPooling2D(pool_size=(2, 2), strides=(2, 2)))

        # Second set of CONV => RELU => POOL layers
        model.add(Conv2D(50, (5, 5), padding='same'))
        model.add(Activation('relu'))
        model.add(MaxPooling2D(pool_size=(2, 2), strides=(2, 2)))

        # First (and only) set of FC => RELU layers
        model.add(Flatten())
        model.add(Dense(500))
        model.add(Activation('relu'))

        model.add(Dense(classes))
        model.add(Activation(last_active))
            
        # return the constructed network architecture
        return model



from .cnn.lenet import LeNet
from .cnn.resnet import ResNet
from .cnn.alexnet import AlexNet
from .cnn.shallownet import ShallowNet 
from .cnn.minivggnet import MiniVGGNet
from .cnn.agegendernet import AgeGenderNet
from .cnn.minigooglenet import MiniGoogLeNet
from .cnn.deepergooglenet import DeeperGoogLeNet

from .meta.siamese import Siamese
from .meta.siameseTripletLost import SiameseTripletLoss
from .meta.reptile import REPTILE

from .data.hdf5DatasetWriter import HDF5DatasetWriter
from .data.hdf5DatasetGenerator import HDF5DatasetGenerator

from .visualize.scatter import point_scatter


# import the necessary packages
from tensorflow.keras.models import Model
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Lambda
from tensorflow.keras.layers import concatenate
from tensorflow.keras.optimizers import Adam

class SiameseTripletLoss:
    
    @staticmethod        
    def triplet_loss(y_true, y_pred):

        dim = y_pred.shape[1] / 3
        anchor, positive, negative = y_pred[ : , 0 : dim ], y_pred[ : , dim : 2 * dim], y_pred[ : , 2 * dim :]
        pos_dist = K.sum(K.square(anchor - positive), axis=-1)
        neg_dist = K.sum(K.square(anchor - negative), axis=-1)
        basic_loss = pos_dist - neg_dist + 0.2
        loss = K.sum(K.maximum(basic_loss, 0.0))
      
        return loss

    @staticmethod
    def build(input_a, input_b, input_c, base_network):
        
        feat_vecs_a = base_network(input_a)
        feat_vecs_b = base_network(input_b)
        feat_vecs_c = base_network(input_c)

        stacked_dists = concatenate([feat_vecs_a, feat_vecs_b, feat_vecs_c], axis=-1)
        outputs = Lambda(lambda x: K.l2_normalize(x, axis=-1))(stacked_dists)
        model = Model(inputs=[input_a, input_b, input_c], outputs=outputs)

        rms = Adam(lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=None, decay=0.0, amsgrad=False)
        model.compile(loss=SiameseTripletLoss.triplet_loss, optimizer=rms)



# import the necessary packages
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import MaxPooling2D
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import Dropout
from tensorflow.keras.layers import Dense
from tensorflow.keras.regularizers import l2
from tensorflow.keras import backend as K

class AlexNet:
    @staticmethod
    def build(width, height, depth, classes, reg=0.0002, last_active="solfmax"):
        # initialize the model along with the input shape to be

        # "channels last" and the channels dimension itself
        model = Sequential()
        inputShape = (height, width, depth)
        chanDim = -1

        # if we are using "channels first", update the input shape
        # and channels dimension
        if K.image_data_format() == "channels_first":
            inputShape = (depth, height, width)
            chanDim = 1

        # Block #1: first CONV => RELU => POOL layer set
        model.add(Conv2D(96, (11, 11), strides=(4, 4),
        input_shape=inputShape, padding="same",
        kernel_regularizer=l2(reg)))
        model.add(Activation("relu"))
        model.add(BatchNormalization(axis=chanDim))
        model.add(MaxPooling2D(pool_size=(3, 3), strides=(2, 2)))
        model.add(Dropout(0.25))

        # Block #2: second CONV => RELU => POOL layer set
        model.add(Conv2D(256, (5, 5), padding="same",
        kernel_regularizer=l2(reg)))
        model.add(Activation("relu"))
        model.add(BatchNormalization(axis=chanDim))
        model.add(MaxPooling2D(pool_size=(3, 3), strides=(2, 2)))
        model.add(Dropout(0.25))

        # Block #3: CONV => RELU => CONV => RELU => CONV => RELU
        model.add(Conv2D(384, (3, 3), padding="same",
        kernel_regularizer=l2(reg)))
        model.add(Activation("relu"))
        model.add(BatchNormalization(axis=chanDim))
        model.add(Conv2D(384, (3, 3), padding="same",
        kernel_regularizer=l2(reg)))
        model.add(Activation("relu"))
        model.add(BatchNormalization(axis=chanDim))
        model.add(Conv2D(256, (3, 3), padding="same",
        kernel_regularizer=l2(reg)))
        model.add(Activation("relu"))
        model.add(BatchNormalization(axis=chanDim))
        model.add(MaxPooling2D(pool_size=(3, 3), strides=(2, 2)))
        model.add(Dropout(0.25))

        # Block #4: first set of FC => RELU layers
        model.add(Flatten())
        model.add(Dense(4096, kernel_regularizer=l2(reg)))
        model.add(Activation("relu"))
        model.add(BatchNormalization())
        model.add(Dropout(0.5))

        # Block #5: second set of FC => RELU layers
        model.add(Dense(4096, kernel_regularizer=l2(reg)))
        model.add(Activation("relu"))
        model.add(BatchNormalization())
        model.add(Dropout(0.5))

        # softmax classifier
        model.add(Dense(classes, kernel_regularizer=l2(reg)))
        model.add(Activation(last_active))

        # return the constructed network architecture
        return model